@@ -51,8 +51,10 @@ use Cache::FileCache;
 use File::Spec;
 use vars qw /$Request/;
 use CGI;
+use Compress::Zlib;
+use Digest::MD5;
 
-our $VERSION        = '0.4';
+our $VERSION = '0.6';
 
 
 sub handler ($$)
@@ -62,7 +64,12 @@ sub handler ($$)
     # Makes $MKDoc::Apache_Cache::Request available
     local $Request = $r;
     
-    my @args = $class->_control_list_process();
+    my @args = do {
+        no warnings;
+        no strict;
+        ( $class->_control_list_process() );
+    };
+
     my ($ret, $data) = $class->_do_cached (@args);
 
     # bug:
@@ -75,6 +82,29 @@ sub handler ($$)
     my ($n_ret) = $data =~ /Status\:\s+(\d+)/;
     $n_ret ||= $ret;
 
+    # if the client doesn't support gzip, fix the headers,
+    # ungzip && send.
+    lc $ENV{'HTTP_ACCEPT_ENCODING'} !~ /gzip/ and
+          $ret =~ /\ncontent-encoding\: gzip/ and do {     
+
+        my ($headers, $body) = split /\r?\n\r?\n/, $data, 2;
+        $headers = join "\r\n",
+                   grep !/content-length\:/i,
+                   grep !/content-encoding\:/i,
+                   grep !/vary\:/i,
+                   split /\r?\n/, $headers;
+        $body = Compress::Zlib::memGunzip ($body);
+        $headers .= "\r\nContent-Length: " . length ($body);
+        $data = $headers . "\r\n\r\n" . $body;
+    };
+
+    $ENV{REQUEST_METHOD} =~ /HEAD/i and do {
+        $data =~ s/\r?\ncontent-length\:.*//i;
+        $data =~ s/\r?\ncontent-encoding\:.*//i;
+        $data =~ s/\r?\nvary\:.*//i;
+        $data =~ s/\r?\netag\:.*//i;
+    };
+
     $r->print ($data);
     return $n_ret;
 }
@@ -90,27 +120,24 @@ sub _do_cached
     $timeout = _expiration_time ($timeout);
     
     my $cache_obj  = $class->_cache_object();
-    my $cached     = $cache_obj->get ($identifier) || do { 
+    my $cached     = $cache_obj->get ($identifier) || do {
+
 	my ($ret, $data) = $class->_do_request();
-	my $tocache = $ret . "\n" . $data;
-        do {
 
-            # add Expires: header to be stored in the cached file
-            my $expires = _http_date ($timeout + time());
-            my $expires_name = $ENV{MKDoc_Apache_Cache_NOEXPIRES} ? 'X-Exp' : 'Expires';
-            $tocache =~ s/\n/\n$expires_name: $expires\n/;
+        # don't cache unless the return code is 200
+        return ($ret, $data) unless ($ret == 200);
 
-            # remove cookies before writing the cached file
-            my $cache_data = $tocache;
-            $ENV{MKDoc_Apache_Cache_NOCOOKIES} and do {
-                $cache_data    =~ s/\nSet-Cookie:.+//g;
-            };
-           
-            $cache_obj->set ($identifier, $cache_data, $timeout);
-        };
-	$tocache;
+        # cookies usually mean non cacheable content
+        return ($ret, $data) if ($data =~ /\nSet-Cookie:.+/);
+
+        # add expires: header to be stored in the cached file
+        my $expires = _http_date ($timeout + time());
+        $data =~ s/\r?\n\r?\n/\r\nExpires: $expires\r\n\r\n/;
+
+        $cache_obj->set ($identifier, "$ret\n$data", $timeout);
+	"$ret\n$data";
     };
-    
+   
     return split /\n/, $cached, 2;
 }
 
@@ -120,7 +147,7 @@ sub _do_request
     my $class  = shift;
     my $fake_r = MKDoc::Apache_Cache::Capture->new ($Request);
     my $ret    = $class->SUPER::handler ($fake_r);
-    return ($ret, $fake_r->data());
+    return ($ret, _make_cache_friendly ($fake_r->data()));
 }
 
 
@@ -177,6 +204,59 @@ sub _cache_object_option
 }
 
 
+# borrowed from http://www.mnot.net/cgi_buffer/ 
+# --------------------------------------------------------------------------------
+# (c) 2000 Copyright Mark Nottingham
+# <mnot@pobox.com>
+#
+# This software may be freely distributed, modified and used,
+# provided that this copyright notice remain intact.
+#
+# This software is provided 'as is' without warranty of any kind.
+#
+# Note from JM: This has been heavily modified from the original, which uses
+# deprecated libs, doesn't compile under 'use strict', and doesn't care about
+# unicode.
+sub _make_cache_friendly
+{
+    my $buf = shift;
+    my ($headers, $body) = split /\r?\n\r?\n/, $buf, 2;
+
+    # we're going to re-write those...
+    $headers =~ s/\r?\ncontent-length\:.*//i;
+    $headers =~ s/\r?\ncontent-encoding\:.*//i;
+    $headers =~ s/\r?\nvary\:.*//i;
+    $headers =~ s/\r?\netag\:.*//i;
+
+    my @o = ();
+
+    # Figure out some kind of content_type
+    my ($content_type) = grep /^content-type\:/i, split (/\r?\n/, $headers);
+    $content_type ||= 'application/octet-stream';
+
+    # Gzip body if content type probably needs gzipping
+ 
+    # Vary: Accept-Encoding is here to tell proxies to keep a separate
+    # cache for every different Accept-Encoding that is being sent.
+    $content_type !~ /zip/ and $content_type =~ /(text|xml)/ and do {
+        $body = Compress::Zlib::memGzip ($body);
+        push @o, "Content-Encoding: gzip";
+        push @o, "Vary: Accept-Encoding";
+    };
+
+    # Compute ETag
+    push @o, "ETag: " . Digest::MD5::md5_hex ($body);
+
+    # Compute Content-Length
+    push @o, "Content-Length: " . length ($body);
+
+    push @o, $headers;
+    push @o, "";
+    push @o, $body;
+    return join "\r\n", @o;
+}
+
+
 # borrowed from http://search.cpan.org/src/RSE/lcwa-1.0.0/lib/lwp/lib/HTTP/Date.pm
 # --------------------------------------------------------------------------------
 our @DoW = qw(Sun Mon Tue Wed Thu Fri Sat);
